/*
* VNC Server for GP2X
*/

#include <stdio.h>
#include <unistd.h>
#include <string.h>
#include <sys/mman.h>
#include <fcntl.h>
#include <linux/fb.h>
#include <signal.h>
#include <time.h>

#include <rfb/rfb.h>

#define WINDOW_WIDTH	320
#define WINDOW_HEIGHT	240
#define WINDOW_DEPTH	2

void sighandler(int signum)
{
  return;
}

int main(int argc,char** argv)
{
	int fbdev;
	unsigned short int *fb,*cmpb, *rfb;
	rfbScreenInfoPtr rfbScreen;
	
	fprintf(stderr,"GP2X VNC SERVER v0.1\n");
	fprintf(stderr," using LibVNCServer 0.7.1\n\n");

	/* Open /dev/fb0 */
	fbdev=open("/dev/fb0", O_RDONLY);
	if(fbdev < 0){
		fprintf(stderr, "DEBUG: Error opening /dev/fb0\n");
		return(-1);
	}


	/* Malloc compare buffer */
	cmpb = (unsigned short int*)malloc(WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH);
	if(cmpb == NULL){
		close(fbdev);
		fprintf(stderr, "DEBUG: Error malloc compare buffer\n");
		return(-1);
	}

	/* Malloc remote buffer */
	rfb = (unsigned short int*)malloc(WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH);
	if(rfb == NULL){
		close(fbdev);
		free(cmpb);
		fprintf(stderr, "DEBUG: Error malloc remote frame buffer\n");
		return(-1);
	}

	/* Mmap /dev/fb */
	fb=(unsigned short int *)mmap(0, WINDOW_WIDTH * WINDOW_HEIGHT * WINDOW_DEPTH, PROT_READ, MAP_SHARED, fbdev, 0);
	if(fb < 0){
		close(fbdev);
		free(cmpb);
		free(rfb);
		fprintf(stderr, "DEBUG: Error mmapping /dev/fb0\n");
		return(-1);
	}
	fprintf(stderr, "DEBUG: fb opened at %x, fd %d\n", fb, fbdev);

	rfbScreen = rfbGetScreen(&argc, argv, WINDOW_WIDTH, WINDOW_HEIGHT, 5, 3,WINDOW_DEPTH);
	rfbScreen->desktopName = "GP2X VNC Server";
	rfbScreen->frameBuffer=(char*)rfb;
	rfbInitServer(rfbScreen);

	/* Make sure we catch EPIPE so we don't get killed by closed sockets when
		the library tries to write to them */
	signal(SIGPIPE,sighandler);

#ifdef FPS_COUNTER
int numframes=0,startclock;
#endif

	while(1){
		while(rfbScreen->clientHead==NULL) {
			rfbProcessEvents(rfbScreen,90000);
		}
		/* Process events */
		rfbProcessEvents(rfbScreen,90000);

#ifdef FULLCOPY

		/* Full Copy mode, copy the whole screen */
		{
			int pos;
			unsigned short int convert,convert2,convert3;
			for(pos=0;pos<WINDOW_WIDTH * WINDOW_HEIGHT;pos++){
				convert = (fb[pos] & (16+8+4+2+1)) << 10;
				convert2 = (fb[pos] & (1024+512+256+128+64)) >> 1;
				convert3 = (fb[pos] & (32768+16384+8192+4096+2048)) >> 11;

				rfb[pos] = convert | convert2 | convert3;
				//rfb[pos] = fb[pos];
			}
			rfbMarkRectAsModified(rfbScreen,0,0,WINDOW_WIDTH,WINDOW_HEIGHT);
		}

#else

#define LOW_B ((fb2[pos] & (16+8+4+2+1)) << 10)
#define LOW_G ((fb2[pos] & (1024+512+256+128+64)) >> 1)
#define LOW_R ((fb2[pos] & (32768+16384+8192+4096+2048)) >> 11)
#define HIGH_B ((fb2[pos] & ((16+8+4+2+1)<<16)) << 10)
#define HIGH_G ((fb2[pos] & ((1024+512+256+128+64)<<16)) >> 1)
#define HIGH_R ((fb2[pos] & ((32768+16384+8192+4096+2048)<<16)) >> 11)

		/* Here we check 2 pixels at once, and even if only one pixel changes we copy them both,
			this speeds up checking for changed pixels. */
		{
			int pos,x1=WINDOW_WIDTH,y1=WINDOW_HEIGHT,x2=-1,y2=-1;
			unsigned int *fb2 = (unsigned int *)fb,
							*cmpb2 = (unsigned int *)cmpb,
							*rfb2 = (unsigned int *)rfb;
			for(pos=0;pos<WINDOW_WIDTH * WINDOW_HEIGHT/2;pos++){
				if(fb2[pos] != cmpb2[pos]){
					int tmp = pos%(WINDOW_WIDTH/2);

					if(x1 > tmp)
						x1 = tmp;

					if(x2 < tmp)
						x2 = tmp;

					tmp = pos/(WINDOW_WIDTH/2);
					if(y1 > tmp)
						y1 = tmp;

					if(y2 < tmp)
						y2 = tmp;

					cmpb2[pos] = fb2[pos];

					rfb2[pos] = HIGH_B | HIGH_G | HIGH_R | LOW_B | LOW_G | LOW_R;
				}
			}

			if(x2 != -1){
				rfbMarkRectAsModified(rfbScreen,x1<<1,y1,(x2<<1)+1,y2+1);
				//fprintf(stderr,"%d: RECT [%d, %d] x [%d, %d]\n",clock()/CLOCKS_PER_SEC,x1<<1,y1,(x2<<1)+1,y2);
			}//else
				//fprintf(stderr,"%d: Nothing changed\n",clock()/CLOCKS_PER_SEC);

#ifdef FPS_COUNTER
			if(numframes%10 == 0){
				if(numframes == 10)
					fprintf(stderr,"FPS: %d\n",CLOCKS_PER_SEC/(clock()-startclock));
				startclock=clock();
				numframes=0;
			}

			numframes++;
#endif

			// Sleep so the actual program we are watching using vnc gets some extra cpu time
			usleep(180000);
		}

#endif
	}

	return(0);
}
